\ TRANSMET.FTH

\ *********************************************************
\ *                                                       *
\ *  F-TP 1.00    [ Turbo-FORTH-83 fr Transputer T800 ]  *
\ *                                                       *
\ *    TRANSMET.FTH - Metacompiler zur Erzeugung eines    *
\ *      32-Bit-Turbo-FORTH-Systems im Transputer-RAM     *
\ *                                                       *
\ *  Fred Behringer,  Planegger Str. 24,   81241 Mnchen  *
\ *                                                       *
\ *                        12.4.1992                      *
\ *                                                       *
\ *********************************************************

\ Als Include-Datei fr das Turbo-FORTH-Hostsystem gedacht.
\ Wird in ein sich dafr im Vokabular  FORTH  einrichtendes
\ Vokabular mit Namen META gelegt.
\ Setzt TRANSSYS.FTH voraus.
\ Metacompiliert die Datei TKERNEL.FTH  oder hnlich aufge-
\ baute,  im Normalfall durch leichte Modifikation  aus der
\ ebengenannten hervorgegangene Dateien  ( auch unter ande-
\ rem Dateinamen ).
\ Das erzeugte Transputer-Forth-System kann ber den Server
\ in TRANSSYS.FTH als COM-Datei abgespeichert werden.
\ Das Ganze ist in der Hauptsache eine fr meine Zwecke um-
\ geschriebene  Version  des Turbo-FORTH-Meta-Compilers der
\ franzsischen FORTH-Gruppe JEDI,  welches wiederum im we-
\ sentlichen der F83-Meta-Compiler von Laxen and Perry ist.
\ Ich habe  zwei Schwierigkeiten:  Der Meta-Assembler  muss
\ auf den Transputer T800 wirken und das FORTH soll ein dem
\ Transputer  angemessenes  32-Bit-FORTH werden.  Ich kenn-
\ zeichne die Stellen, die ich angepasst habe, mit \ ***

\ Mit INCLUDE TRANSMET wird der Meta-Compiler ins Host-
\ System compiliert.  Dabei wird  mit  den Mitteln  des
\ Host-Systems gearbeitet.Insbesondere wird unterschie-
\ den zwischen 16-Bit-Zahlen  ( normal,  compiliert mit
\ (LIT) ) und 32-Bit-Zahlen  ( doppeltgenau, compiliert
\ mit (LIT) (LIT) ).

\ Der Meta-Compiler ist dazu da,  TKERNEL.FTH ins Ziel-
\ system zu "compilieren", also das Zielsystem zu gene-
\ rieren.

\ Der Meta-Compiler  akzeptiert  nur Zahlen,  die "ein-
\ fach-genau",  also ohne Punkt  geschrieben sind.  Das
\ sind aber 32-Bit-Zahlen  ( einfach-genau im Sinne des
\ Zielsystems ).  Der Meta-Compiler  interpretiert also
\ Zahlen  aus TKERNEL.FTH,  die ohne Punkt  geschrieben
\ wurden,  als 32-Bit-Zahlen  und compiliert  diese als
\ solche ins Zielsystem.  Zahlen mit Punkt,  also "dop-
\ pelt-genaue" Zahlen des Zielsystems, d.h. 64-Bit-Zah-
\ len,  werden erst  beim Nachladen von  TCORE.FTH  vom
\ dann schon bestehenden Zielsystem-Kern akzeptiert.

\ Nach dem Laden des Meta-Compilers  ber INCLUDE TRANSMET
\ wird ber INCLUDE TKERNEL der Zielfunktions-Kern TKERNEL
\ geladen und metacompiliert. Da fr den Meta-Compiler der
\ normale Interpreter des Hostsystems als Interpreter mit-
\ fungiert,  wird das Wort CONVERT des Host-Systems am An-
\ fang  von TKERNEL.FTH  so umgeschaltet, dass whrend der
\ ganzen Metacompilation punktfrei geschriebene Zahlen als
\ 32-Bit-Zahlen interpretiert  und Punktzahlen nicht aner-
\ kannt werden.  CONVERT  wird dazu "gepatcht".  Um dieses
\ Umschalten  nach Beendigung  der Metacompilation  sauber
\ wieder rckgngig machen zu knnen, wird die ursprngli-
\ che cfa  von CONVERT  des Host-Systems  in der Variablen
\ PSAVE aufbewahrt. Fr das Zurckpatchen sorgt das am En-
\ de von TKERNEL.FTH direkt wirkende Wort END-META mit.

DECIMAL

\ Vokabulare einrichten

ONLY FORTH ALSO  DEFINITIONS
VOCABULARY META
: [META]      META                   ; IMMEDIATE
META ALSO  META DEFINITIONS
: [ONLY]      ONLY                   ; IMMEDIATE
: [FORTH]     FORTH                  ; IMMEDIATE
: [META]      META                   ; IMMEDIATE
: [ASSEMBLER] [COMPILE] [TASSEMBLER] ; IMMEDIATE
: [SERVER]    [COMPILE] [SERVER]     ; IMMEDIATE

: SWITCH ( -- )  NOOP ( Context )  NOOP ( Current )
   DOES>  DUP @  CONTEXT @  SWAP CONTEXT !  OVER ! 2+
          DUP @  CURRENT @  SWAP CURRENT !  SWAP ! ;
   \ Speichert den Inhalt von pfa nach CONTEXT und den von pfa+2
   \ nach CURRENT.  Dann ldt es den Inhalt  von CONTEXT und den
   \ von CURRENT eben dorthin und bewahrt diese auf.

SWITCH   ( bereitet sich selbst auf seine Runtime-Prozedur vor )

VOCABULARY TARGET   VOCABULARY TRANSITION
VOCABULARY FORWARD  VOCABULARY USER
ONLY DEFINITIONS FORTH ALSO META ALSO
: META       META       ;  : TARGET  TARGET  ;
: TRANSITION TRANSITION ;  : FORWARD FORWARD ;
: USER       USER       ;  : SERVER  SERVER  ;
: ASSEMBLER  TASSEMBLER ;

ONLY FORTH ALSO TASSEMBLER ALSO DEFINITIONS
: META       META       ;

ONLY FORTH ALSO META ALSO DEFINITIONS
: ASSEMBLER  TASSEMBLER ;
: IN-TARGET     ( -- ) ONLY TARGET DEFINITIONS ;
: IN-TRANSITION ( -- )
        ONLY FORWARD ALSO TARGET DEFINITIONS ALSO TRANSITION ;
: IN-META       ( -- ) ONLY FORTH ALSO META DEFINITIONS ALSO ;
: IN-FORWARD    ( -- ) FORWARD DEFINITIONS ;

: -FIND ( <wort> -- cfa f ) \ Suche in nur einem einzigen Vokabular
   BL WORD ?UPPERCASE DUP CONTEXT @ HASH @ (FIND) ;

\ Zugriff zum TRAM von META aus

IN-META SERVER ALSO

: C@-T C@-T ;  : @-T  @-T ;  : DP-T DP-T ;
: C!-T C!-T ;  : !-T  !-T ;  : C,-T C,-T ;
: ALIGN-T ALIGN-T ;
: CSET-T  ( o ad. -- ) ROT >R 2DUP C@-T R> OR -ROT C!-T ;
: HERE-T  ( -- ad. )   DP-T  2@ ;
: ALLOT-T ( d -- )     DP-T D+! ;
:  ,-T    ( d -- )     HERE-T       !-T  4. ALLOT-T ;
: S,-T    ( ad l -- )  0  ?DO COUNT C,-T LOOP  DROP ;

META IN-META
\ Variablen und Umschaltmglichkeiten fr den Metacompiler
 VARIABLE      WIDTH     31 WIDTH !  \ Wortkpfe: Lnge der nfa
2VARIABLE      LAST-T                \ zuletzt metacompilierte nfa
2VARIABLE      CONTEXT-T             \ Meta-CONTEXT-Vokabular
2VARIABLE      CURRENT-T             \ Meta-CURRENT-Vokabular
 VARIABLE      CHECKING              \ bedingte Compilation
 VARIABLE      FORCE                 \ erzwungene Compilation
 VARIABLE      VALID     VALID ON    \ Gltigerklrung der gerade
                                     \ bearbeiteten Definition
2VARIABLE      OLD-T                 \ ursprngliche Position im T-System
 VARIABLE      SAVING    SAVING ON   \ mit oder ohne Abspeichern
32 STRING      TARGET$               \ Dateiname fr das
" TFORTH.FYS"  TARGET$ $!            \ abzuspeichernde T-System
: .FLAG   ( ad -- )
  @ ?DUP IF ." ON " DUP -1 <> IF ." ( " . ." )" ELSE DROP THEN
         ELSE ." OFF" THEN ;
: OPTIONS ( -- )
  CR ." Wahlmglichkeiten :"
  CR ."   Wortkopflnge                   WIDTH " WIDTH    .FLAG
  CR ."   Bedingte Compilation         CHECKING " CHECKING .FLAG
  CR ;

\ Worte zur Umschaltung auf 32-Bit-Organisation
: UNERLAUBT TRUE ABORT" Zur Meta-Compilation nicht zugelassen" CR ;
2VARIABLE SAVE-CONVERT
: (CONVERT) ( +d1 adr1 -- +d2 adr2 )
   BEGIN  1+  DUP >R  C@  BASE @  DIGIT
   WHILE  SWAP  BASE @ UM*  DROP  ROT  BASE @ UM*  D+
   DOUBLE?  IF CR ." Punktzahlen " UNERLAUBT THEN  R>
   REPEAT
   DOUBLE?  IF CR ." Punktzahlen " UNERLAUBT THEN  DROP  R>  0 DPL ! ;
: PATCH-CONVERT ( -- )
   ['] CONVERT >R R@ 2+ 2@ SAVE-CONVERT 2!
   ['] (CONVERT) R@ 2+ ! ['] EXIT R> 4+ ! ;
: REPATCH-CONVERT ( -- )
   SAVE-CONVERT 2@ ['] CONVERT 2+ 2! ;

\ Compilation und Auflsung der Referenzen in TARGET und FORWARD
: MAKE-CODE ( pfa -- )
   DUP 4+ @ IF 2@ ELSE DROP 0. THEN ,-T ;  \ *** 4+ 2@ 0.
: LINK-BACKWARDS ( pfa -- )
   VALID @ IF 0 HERE-T 2OVER DROP 2@ ,-T 2SWAP DROP 2!
           ELSE DROP 0. ,-T THEN ;      \ *** 0 2OVER DROP 2@ 2SWAP DROP 2! 0.
: RESOLVED? ( pfa -- f ) 4+ @ ;         \ *** 4+
: FORWARD-CODE ( pfa -- )
   DUP RESOLVED? IF MAKE-CODE ELSE LINK-BACKWARDS THEN ;
: FORWARD: ( -- )
   SWITCH FORWARD DEFINITIONS CREATE SWITCH 0 , 0 , 0 ,
   DOES> FORWARD-CODE ;                 \ *** 0 ,
: 'T ( -- cfa ) CONTEXT @ TARGET DEFINED ROT CONTEXT ! 0= ?MISSING ;
: [TARGET]   ( -- )   'T , ; IMMEDIATE  \ *** Nachtrglich wieder aufgenommen
                                        \ *** War mit \ ausgeschaltet
: 'F ( -- cfa f ) CONTEXT @ FORWARD -FIND ROT CONTEXT ! ;
: [FORWARD] ( -- ) 'F 0= ?MISSING , ; IMMEDIATE
: RESOLVE ( ad. cfa -- )
   >BODY 3DUP TRUE OVER 4+ ! 2@
   BEGIN 2DUP D0<>
   WHILE 4DUP @-T 2SWAP 2ROT !-T
   REPEAT 4DROP 2! ;    \ *** 3DUP 4+ 2@ 2DUP D0<> 4DUP 2SWAP 2ROT 4DROP 2!
: ?RESOLVES ( -- )
   'F IF DUP >BODY RESOLVED? 0= IF HERE-T ROT RESOLVE ELSE DROP THEN
      ELSE DROP THEN ;  \ *** ROT

\ Frgltigerklrung der Definitionen bei bedingter Meta-Compilation
DEFER ?VALID ( -- )     \ regelt die Gltigkeit eines neuen Wortes
: (?VALID) ( -- )       \ Frgltigerklrung von "markierten" Worten in FORTH
   CHECKING @ IF
   >IN @ CONTEXT @
   FORTH -FIND IF >VIEW @ 0<> ELSE DROP FALSE THEN VALID !
   CONTEXT ! >IN ! THEN ;
' (?VALID) IS ?VALID
: ?BACKSTEP ( -- )
   VALID @ IF HERE-T OLD-T 2! ELSE OLD-T 2@ DP-T 2! THEN ;
                    \ *** 2! 2@ 2!
: ?HERE-T ( -- ad. )
   VALID @ IF HERE-T ELSE 0. THEN ;  \ *** 0.
: FORCED ( -- )  \ unabdingbare Compilation
   ?BACKSTEP  CHECKING @  FORCE !  CHECKING OFF  VALID ON ;
: UNFORCED ( -- )  \ Ende der unabdingbaren Compilation
   FORCE @  CHECKING ! ;

\ Sprungmarken zu Codeabschnitten im TRAM; werden in TASSEMBLER aufbewahrt
: LABEL ( -- )
   FORCED  >IN @  ASSEMBLER DEFINITIONS  HERE-T 2CONSTANT  >IN !
   ?RESOLVES UNFORCED ;  \ *** 2CONSTANT

\ Erzeugung von Wortkpfen im TRAM
: HASH ( str-ad voc-ad. -- thread. )
   ROT 1+ C@ #THREADS 1- AND 4* 0 D+ ;  \ *** ROT 4* 0 D+ ;
: HEADER ( -- )
   BL WORD  C@ 1+  WIDTH @  MIN  DUP  VALID @ AND
   IF ALIGN-T 0. ,-T ( vfa ) HERE CURRENT-T 2@ HASH 2DUP @-T ,-T
     HERE-T D4- 2SWAP !-T HERE-T HERE -ROT 2SWAP SWAP
     DUP >R                 \ Lnge des Namens aufbewahren; nderung 7.9.95
     S,-T
     128 HERE-T D1- CSET-T
     R> 1- C,-T             \ Extrabyte mit Lnge erzeugen; nderung 7.9.95
     ALIGN-T 2DUP LAST-T 2! 128 -ROT CSET-T
   ELSE DROP THEN  ;
   \ *** ALIGN-T 0. 2@ 2DUP D4- 2SWAP -ROT 2SWAP SWAP ALIGN-T 2DUP 2! -ROT D1-

\ Erzeugung im TRAM und in TARGET mit automatischer Auflsung ber FORWARD
: TARGET-CREATE ( -- )
   ?BACKSTEP ?VALID >IN @ DUP HEADER >IN !
   IN-TARGET CREATE IN-META  HERE-T , , VALID @ ,
   >IN ! ?RESOLVES  DOES> MAKE-CODE ;  \ *** ,
: RECREATE ( -- ) >IN @ TARGET-CREATE >IN ! ;
: CODE ( -- ) TARGET-CREATE HERE-T D4+ ,-T ASSEMBLER !CSP ;  \ *** D4+

ASSEMBLER ALSO DEFINITIONS
: END-CODE IN-META ?CSP ;
: C; END-CODE ;
META IN-META

\ Compilation der Worte fr TRANSITION
: T: ( -- )
   SWITCH  TRANSITION DEFINITIONS  CREATE  SWITCH  ]  DOES> >R ;  \ >R ???
: T; ( -- )
   SWITCH  TRANSITION DEFINITIONS  [COMPILE] ;  SWITCH ; IMMEDIATE
: DIGIT? ( char -- f ) BASE @ DIGIT NIP ;
: PUNCT? ( char -- f )
   ASCII . OVER = SWAP ASCII - OVER = SWAP ASCII / OVER = SWAP DROP OR OR ;
: NUMERIC? ( ad len -- f )
   DUP 1 = IF DROP C@ DIGIT? EXIT THEN
   1 -ROT 0 ?DO DUP C@ DUP DIGIT? SWAP PUNCT? OR
   ROT AND SWAP 1+ LOOP DROP ;

\ Worte fr TRANSITION
T: (  [COMPILE] (  T;
T: (S [COMPILE] (S T;
T: \  [COMPILE] \  T;

: ]]  ] ;
: [[  [COMPILE] [ ; FORTH IMMEDIATE META
FORWARD: DEFINITIONS
FORWARD: [

: STRING,-T ( -- ) ASCII " PARSE DUP C,-T S,-T ALIGN-T ;  \ *** ALIGN-T
FORWARD: (.")
T: ." [FORWARD] (.") STRING,-T T;
FORWARD: (")
T: "  [FORWARD] (")  STRING,-T T;
FORWARD: (ABORT")
T: ABORT" [FORWARD] (ABORT") STRING,-T T;
FORWARD: (LIT)
T: LITERAL ( n-t -- )  [FORWARD] (LIT) ,-T T;  \ ***  okay ?
T: DLITERAL ( "d" -- ) [FORWARD] (LIT) ,-T T;  \ *** okay ?
\ T: DLITERAL ( "d" -- ) [FORWARD] (LIT) ,-T [FORWARD] (LIT) ,-T  T;
T: ASCII ( -- ) [COMPILE] ASCII 0 [[ TRANSITION ]] LITERAL [META] T;

\ Definierende Worte
FORWARD: DOCREATE
: CREATE RECREATE [FORWARD] DOCREATE ?HERE-T 2CONSTANT ;  \ *** 2CONSTANT
: VARIABLE ( -- ) CREATE 0. ,-T ;
FORWARD: DODEFER
: DEFER ( -- ) TARGET-CREATE [FORWARD] DODEFER 0. ,-T ;   \ *** 0.
FORWARD: DOCONSTANT
: CONSTANT ( n-T -- )
   RECREATE [FORWARD] DOCONSTANT 2DUP ,-T 2CONSTANT ;     \ *** 2DUP 2CONSTANT
\ FORWARD: DO2CONSTANT
\ : 2CONSTANT     ( d -- )
\      RECREATE   [FORWARD] DO2CONSTANT
\                 OVER OVER  ,-T  ,-T  2CONSTANT ;
FORWARD: DOSTRING
: STRING ( n-T -- )
   RECREATE [FORWARD] DOSTRING
   DROP DUP C,-T 0 C,-T DUP 0 ALLOT-T STRING ;
FORTH 2VARIABLE #USER-T META ALSO USER DEFINITIONS        \ *** 2VARIABLE
: ALLOT ( n-T -- ) #USER-T D+! ;                          \ *** D+!
FORWARD: DOUSER-VARIABLE
: VARIABLE ( -- )
   SWITCH RECREATE [FORWARD] DOUSER-VARIABLE #USER-T 2@
   2DUP ,-T 4. ALLOT META DEFINITIONS
   [FORTH] 2CONSTANT [META] SWITCH ;                 \ *** 2@ 2DUP 2CONSTANT 4.
FORWARD: DOUSER-DEFER
: DEFER    ( -- )
   SWITCH RECREATE [FORWARD] DOUSER-DEFER    #USER-T 2@
   2DUP ,-T 4. ALLOT META DEFINITIONS
   [FORTH] 2CONSTANT [META] SWITCH ;                 \ *** 2@ 2DUP 2CONSTANT 4.
META IN-META

\ berleitungsworte ( TRANSITION ) und Vokabulare
FORTH  2VARIABLE VOC-LINK-T  META                    \ *** 2VARIABLE
FORWARD: <VOCABULARY>
: VOCABULARY ( -- )
   RECREATE [FORWARD] <VOCABULARY> HERE-T #THREADS 0 DO 0. ,-T LOOP
   HERE-T  VOC-LINK-T 2@ ,-T  VOC-LINK-T 2!  [FORTH] 2CONSTANT
   DOES> 2@ [META] CONTEXT-T 2! ;             \ *** 0. 2@ 2! 2CONSTANT 2@ 2!

FORWARD: (;USES)
FORTH  VARIABLE STATE-T  META
T: ;USES ( -- )
     [FORWARD] (;USES)  IN-META ASSEMBLER  !CSP  STATE-T OFF  T;
T: [COMPILE]  'T EXECUTE  T;
FORWARD: (IS)
T: IS  [FORWARD] (IS) T;
: IS 'T >BODY DUP 4+ @ IF 2@ D4+ ( >BODY-T ) !-T ELSE DROP 2DROP THEN ;
                                                     \ *** 4+ 2@ D4+ 3DROP
T: ALIGN  ALIGN-T  T;
\ T: EVEN T;  \ Brauche ich das zum Metacompilieren ???
FORWARD: (;CODE)
FORWARD: DODOES
T: DOES> ( <forward-type> -- )
    [FORWARD] (;CODE)  ?RESOLVES
    [FORWARD] DODOES
    HERE-T D4- DP-T 2!
    HERE-T @-T
    SWITCH [ASSEMBLER]
    LDC  0. ADC  0. ADC  0. ADC  GCALL        \ Die ADCs sind zum Wortausgleich
    SWITCH T;

: NUMERIC ( -- )
    [FORTH] HERE [META] NUMBER DPL @ 1+
    IF [[ TRANSITION ]] DLITERAL [META]
    ELSE DROP [[ TRANSITION ]] LITERAL [META] THEN ;
: UNDEFINED ( -- )
    HERE-T 0. ,-T IN-FORWARD [FORTH]
    CREATE [META] TRANSITION [FORTH] , FALSE , [META]
    DOES> FORWARD-CODE ;  \ *** 0.

\ Compilations-Schleife des Meta-Generators
[FORTH] VARIABLE T-IN  META
: ] ( -- )
    STATE-T ON  IN-TRANSITION
    BEGIN  >IN @ T-IN !  DEFINED
      IF   EXECUTE
      ELSE COUNT NUMERIC?
           IF NUMERIC
           ELSE  VALID @  IF  T-IN @ >IN !  UNDEFINED  THEN
           THEN
      THEN STATE-T @ 0= END? @ OR END? OFF
     UNTIL ;
: NEWSTATE STATE-T @ IF ] THEN STATE ;
[FORTH]  ' NEWSTATE IS RUN  META
T: [ ( -- ) IN-META  STATE-T OFF  T;
FORWARD: UNNEST
T: ; ( -- ) [FORWARD] UNNEST [[ TRANSITION ]] [  T;
FORWARD: NEST

\ Steuerelemente von META
: 4SWAP  7 ROLL 7 ROLL 7 ROLL 7 ROLL ;
: ?CONDITION ( f. -- ) D0= ABORT" Struktur unvollstaendig" ;  \ *** neu
: ?>MARK     ( -- f. Tad ) -1. HERE-T 0. ,-T ;                \ *** -1. 0.
: ?>RESOLVE  ( f. Tad -- ) HERE-T 2SWAP !-T ?CONDITION ;      \ *** 2SWAP
: ?<MARK     ( -- f. Tad ) -1. HERE-T ;                       \ *** -1.
: ?<RESOLVE  ( f. Tad -- ) ,-T ?CONDITION ;
FORWARD: BRANCH
FORWARD: ?BRANCH
T: BEGIN  ?<MARK    T;
T: AGAIN  [FORWARD] BRANCH  ?<RESOLVE  T;
T: UNTIL  [FORWARD] ?BRANCH ?<RESOLVE  T;
T: IF     [FORWARD] ?BRANCH ?>MARK     T;
T: THEN   ?>RESOLVE T;
T: ELSE   [FORWARD] BRANCH  ?>MARK 4SWAP ?>RESOLVE T;         \ *** 4SWAP
T: WHILE  [[ TRANSITION ]]  IF  T;
T: REPEAT 4SWAP [[ TRANSITION ]] AGAIN THEN T;                \ *** 4SWAP
FORWARD: (?DO)
FORWARD: (DO)
FORWARD: (LOOP)
FORWARD: (+LOOP)
FORWARD: (LEAVE)
FORWARD: (?LEAVE)
T: ?DO    [FORWARD] (?DO)    ?>MARK T;
T: DO     [FORWARD] (DO)     ?>MARK T;
T: LOOP   [FORWARD] (LOOP)   4DUP D4+ ?<RESOLVE ?>RESOLVE T;  \ *** 4DUP D4+
T: +LOOP  [FORWARD] (+LOOP)  4DUP D4+ ?<RESOLVE ?>RESOLVE T;  \ *** 4DUP D4+
T: LEAVE  [FORWARD] (LEAVE)  T;
T: ?LEAVE [FORWARD] (?LEAVE) T;
FORWARD: OVER                                                 \ *** ???
FORWARD: DROP                                                 \ *** ???
FORWARD: =                                                    \ *** ???
T: CASE    CSP @ !CSP T;                                      \ *** ???
T: OF      [FORWARD] OVER    [FORWARD] =
           [FORWARD] ?BRANCH ?>MARK
           [FORWARD] DROP    T;                               \ *** ???
T: ENDOF   [[ TRANSITION ]] ELSE T;
T: ENDCASE [FORWARD] DROP
           [FORTH] BEGIN SP@ CSP @ <> WHILE [META] ?>RESOLVE
           [FORTH] REPEAT CSP ! [META] T;                     \ *** ???

\ Schlussbilanz der Meta-Compilation und Abspeicherung
: .SYMBOLS ( -- ) \ Liste der fr gltig erklrten TARGET-Symbole
     CR ." Liste der Worte des T-Systems :" CR
     TARGET CONTEXT @ HERE #THREADS 2* CMOVE
     BEGIN HERE #THREADS LARGEST DUP
     WHILE DUP L>NAME  DUP C@  31 AND #OUT @ + RMARGIN @ >
          IF  CR LMARGIN @ SPACES  THEN
          DUP NAME> >BODY 4+ @ IF .ID  ELSE DROP THEN
          @ SWAP !  STOP?
          IF EXIT THEN
     REPEAT 2DROP IN-META ;                                   \ *** 4+
: .UNRESOLVED ( -- )
     FORWARD  CONTEXT @ HERE #THREADS 2* CMOVE
     BEGIN HERE #THREADS LARGEST DUP
     WHILE ?CR DUP L>NAME NAME> >BODY RESOLVED? 0=
          IF  DUP L>NAME .ID  THEN
           @  SWAP !  STOP?
          IF 2DROP IN-META EXIT THEN
     REPEAT  2DROP IN-META ;
HEX

: TSAVE ( -- )
CR BOLD ." Zum Einladen des zweiten Systemteils entweder jetzt eingeben:"
CR      ." SERVER TRESET GOTO-SYSTEM [RET] ECHO OFF INCLUDE TCORE [RET]"
CR      ." und das Gesamtsystem dann abspeichern mit:"
CR      ." SAVE-SYSTEM TFORTH.FYS [RET]"
CR      ." oder zunchst das erste Teilsystem abspeichern mit:"
CR      ." SERVER TRESET GOTO-SYSTEM SAVE-SYSTEM TFORTH.FYS [RET],"
CR      ." dann auf die DOS-Ebene gehen mit:"
CR      ." [ESC] BYE [RET]"
CR      ." und von dort aus spter das zweite Teilsystem hinzuladen mit:"
CR      ." LOAD-FTP [RET] INCLUDE TCORE [RET]"
CR      ." und dann erst das Gesamtsystem abspeichern mit"
CR      ." SAVE-SYSTEM TFORTH.FYS ."
CR      ." Natrlich kann man statt TFORTH.FYS auch andere Namen verwenden"
CR      ." und damit z.B. zwischen einem vorcompilierten ersten Teilsystem"
CR      ." und einem vorcompilierten Gesamtsystem unterscheiden ( sie also"
CR      ." beide  fr eventuelle Experimente bereitstellen ). Man muss das"
CR      ." dann jedoch geeignet bercksichtigen, wenn man zum Einladen des"
CR      ." Teil oder Gesamtsystems die den Vorgang automatisierende Batch-"
CR      ." datei LOAD-FTP.BAT verwenden mchte."
CR CR   ." Weiter: [Taste]" ATTOFF
KEY 0= IF KEY DROP THEN ;

: STATISTIK ( -- )
  CR BOLD ." Nicht aufgelste Vorwrtsreferenzen:" CR .UNRESOLVED CR
  CR ." BILANZ: "
  CR ." Host-Segment:                   " HEX DSEGMENT 0A U.R
  CR ." Letzte Host-Adresse:            " [FORTH] HERE 0A U.R
  CR ." Erste  T-System-Adresse   WPTR: " [SERVER] HEX WPTR 0A UD.R
  CR ."                            ORG: " ORG 2@ 0A UD.R
  CR ." Letzte T-System-Adresse HERE-T: " HERE-T 0A UD.R
  CR ."                            EOS: " EOS 2@ 0A UD.R DECIMAL CR
  CR ." Lnge des T-Systems eigentlich: " HERE-T WPTR D- 0A UD.R ."  d"
  CR ."                      insgesamt: " EOS 2@ WPTR D- 0A UD.R ."  d" CR
  CR ." Weiter: [Taste]" ATTOFF
  KEY 0= IF KEY DROP THEN ;

\ Umschalten auf 32-Bit-Organisation
: 2VARIABLE CR ." 2VARIABLE" UNERLAUBT ;

: START-META ( -- ) PATCH-CONVERT ;
: END-META   ( -- )
  STATISTIK
  DMA 3 + C@ 20 <> IF TSAVE THEN   \ Interpretation siehe unten
  REPATCH-CONVERT
  [FORTH] ONLY FORTH ALSO DEFINITIONS SERVER ;

\ TRANSMET INCLUDE TKERNEL   ...  => Server- und Metacompilermeldungen
\ TRANSMET  INCLUDE TKERNEL  ...  => Metacompiler-, aber keine Servermeldung
\ TRANSMET   INCLUDE TKERNEL ...  => Keine Meldungen
\ TRANSMET 0 INCLUDE TKERNEL ...  => Server-, aber keine Metacompilermeldung

\ Letzte Interpretations-Anweisungen an den Meta-Compiler
 : H:  [COMPILE] :  ;
 : H;  [COMPILE] ;  ; IMMEDIATE
H: 0 0. H;
H: 1 1. H;
H: 2 2. H;
H: 3 3. H;
H: '  'T >BODY 2@ H;                                          \ *** 2@
H: ,  ,-T         H;
H: C, C,-T        H;
H: HERE HERE-T    H;
H: ALLOT ALLOT-T  H;
H: DEFINITIONS DEFINITIONS  CONTEXT-T 2@  CURRENT-T 2!  H;    \ *** 2@ 2!
DECIMAL
H: IMMEDIATE ( -- ) WIDTH @ IF ( Headers ? )
         64 ( Precedence Bit ) LAST-T 2@  CSET-T  THEN  H;    \ *** 2@
H: : ( -- ) TARGET-CREATE [FORWARD] NEST ]              H;

\ Seltsam. Das scheint wirklich die letzte Anweisung an den Meta-Compiler
\ sein zu mssen.  Der Doppelpunkt  gibt in dem zu erzeugenden Zielsystem
\ die Anweisung,  die darauf folgenden Worte  ( des Zielsystems ) bis zum
\ abschlieenden Semikolon zu compilieren.  Der Doppelpunkt hier gibt dem
\ Meta-Compiler die Anweisung, diese Anweisung metacompilierend auszufh-
\ ren.

\ Beim Compilieren des Meta-Compilers wird SWITCH ( als Doppelpunktwort )
\ definiert und unmittelbar darauf direkt ausgefhrt. Eine solche direkte
\ Ausfhrung  eines gerade definierten Wortes  ist natrlich  im gesamten
\ TKERNEL nicht mglich,  da  das metazucompilierende System  whrend der
\ Meta-Compilation ja noch gar nicht arbeitet. Direkt auszufhrende Worte
\ ( die  also  nicht  in Doppelpunkt- oder Codedefinitionen  stecken ) in
\ TKERNEL mssen im Hostsystem im gerade eingeschalteten Vokabular  ( zu-
\ meist wohl META )  definiert sein  und werden,  wenn  vom Meta-Compiler
\ nicht anders vorgesehen  ( wie bei  VARIABLE  als  definierendem Wort )
\ whrend der Metacompilation ausgefhrt.

BOLD CR
.(   Meta-Compiler geladen  )
ATTOFF META OPTIONS FORTH BOLD
.( )
CR CR ATTOFF
[ONLY] FORTH ALSO DEFINITIONS SERVER
HEX


eof
-----------------------------------------------------------------
